"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = exports.ERROR_DB_LOCKED = void 0;
var _debug = _interopRequireDefault(require("debug"));
var _lodash = _interopRequireDefault(require("lodash"));
var _lowdb = _interopRequireDefault(require("lowdb"));
var _FileAsync = _interopRequireDefault(require("lowdb/adapters/FileAsync"));
var _Memory = _interopRequireDefault(require("lowdb/adapters/Memory"));
var _path = _interopRequireDefault(require("path"));
var _core = require("@verdaccio/core");
var _utils = require("@verdaccio/utils");
var _dirUtils = require("./dir-utils");
var _fs = require("./fs");
var _localFs = _interopRequireWildcard(require("./local-fs"));
var _pkgUtils = require("./pkg-utils");
var _utils2 = require("./utils");
var _process$env$VERDACCI; // import LRU from 'lru-cache';
function _getRequireWildcardCache(e) { if ("function" != typeof WeakMap) return null; var r = new WeakMap(), t = new WeakMap(); return (_getRequireWildcardCache = function (e) { return e ? t : r; })(e); }
function _interopRequireWildcard(e, r) { if (!r && e && e.__esModule) return e; if (null === e || "object" != typeof e && "function" != typeof e) return { default: e }; var t = _getRequireWildcardCache(r); if (t && t.has(e)) return t.get(e); var n = { __proto__: null }, a = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var u in e) if ("default" !== u && {}.hasOwnProperty.call(e, u)) { var i = a ? Object.getOwnPropertyDescriptor(e, u) : null; i && (i.get || i.set) ? Object.defineProperty(n, u, i) : n[u] = e[u]; } return n.default = e, t && t.set(e, n), n; }
function _interopRequireDefault(e) { return e && e.__esModule ? e : { default: e }; }
const TOKEN_DB_NAME = '.token-db.json';
const DB_NAME = (_process$env$VERDACCI = process.env.VERDACCIO_STORAGE_NAME) !== null && _process$env$VERDACCI !== void 0 ? _process$env$VERDACCI : _core.fileUtils.Files.DatabaseName;
const debug = (0, _debug.default)('verdaccio:plugin:local-storage');
const ERROR_DB_LOCKED = exports.ERROR_DB_LOCKED = 'Database is locked, please check error message printed during startup to prevent data loss';
class LocalDatabase extends _core.pluginUtils.Plugin {
  constructor(config, logger) {
    // TODO: fix double config
    super(config, {
      config,
      logger
    });
    this.tokenDb = null;
    this.config = config;
    this.logger = logger;
    this.locked = false;
    this.data = undefined;
    debug('config path %o', config.configPath);
    this.path = (0, _utils2._dbGenPath)(DB_NAME, config);
    this.storages = this._getCustomPackageLocalStorages();
    this.logger.info({
      path: this.path
    }, 'local storage path @{path}');
    debug('plugin storage path %o', this.path);
  }
  async init() {
    debug('plugin init');
    this.data = await this.fetchLocalPackages();
    debug('local packages loaded');
    await this._sync();
  }
  async getSecret() {
    if (typeof this.data === 'undefined') {
      throw Error('no data secret available');
    }
    return Promise.resolve(this.data.secret);
  }
  async setSecret(secret) {
    if (typeof this.data === 'undefined') {
      throw Error('no data secret available');
    } else {
      this.data.secret = secret;
    }
    await this._sync();
  }
  async add(name) {
    if (typeof this.data === 'undefined') {
      throw Error('no data secret available');
    }
    if (this.data.list.indexOf(name) === -1) {
      this.data.list.push(name);
      debug('the private package %s has been added', name);
      await this._sync();
    } else {
      debug('the private package %s already exist on database', name);
      return Promise.resolve();
    }
  }

  /**
   * The field storage could be absolute or relative.
   * If relative, it will be resolved against the config path.
   * If absolute, it will be returned as is.
   **/
  getStoragePath() {
    const {
      storage
    } = this.config;
    if (typeof storage !== 'string') {
      throw new TypeError('storage field is mandatory');
    }
    const storagePath = _path.default.isAbsolute(storage) ? storage : _path.default.normalize(_path.default.join(this.getBaseConfigPath(), storage));
    debug('storage path %o', storagePath);
    return storagePath;
  }
  getBaseConfigPath() {
    return _path.default.dirname(this.config.configPath);
  }

  /**
   * Filter and only match those values that the query define.
   **/
  async filterByQuery(results, query) {
    // FUTURE: apply new filters, keyword, version, ...
    return results.filter(item => {
      var _item$name;
      return (item === null || item === void 0 || (_item$name = item.name) === null || _item$name === void 0 ? void 0 : _item$name.match(query.text)) !== null;
    });
  }

  // eslint-disable-next-line @typescript-eslint/no-unused-vars
  async getScore(_pkg) {
    // TODO: there is no particular reason to predefined scores
    // could be improved by using
    return Promise.resolve({
      final: 1,
      detail: {
        maintenance: 0,
        popularity: 1,
        quality: 1
      }
    });
  }

  /**
   *
   * @param query
   * @returns
   */
  async search(query) {
    debug('search query to %o', query.text);
    const results = [];
    const storagePath = this.getStoragePath();
    const localResults = await (0, _dirUtils.searchOnStorage)(storagePath, this.storages);
    const packagesOnStorage = await this.filterByQuery(localResults, query);
    debug('packages found %o', packagesOnStorage.length);
    for (let storage of packagesOnStorage) {
      // check if package is listed on the cache private database
      const isPrivate = this.data.list.includes(storage.name);
      const score = await this.getScore(storage);
      results.push({
        package: storage,
        verdaccioPrivate: isPrivate,
        verdaccioPkgCached: !isPrivate,
        score
      });
    }
    return results;
  }
  async remove(name) {
    try {
      if (typeof this.data === 'undefined') {
        throw Error('no data secret available');
      }
      const data = await this.get();
      const pkgName = data.indexOf(name);
      if (pkgName !== -1) {
        this.data.list.splice(pkgName, 1);
        debug('remove package %o has been removed', name);
      }
      await this._sync();
    } catch (err) {
      this.logger.error({
        err
      }, 'remove the private package has failed @{err}');
      throw _core.errorUtils.getInternalError('error remove private package');
    }
  }
  async get() {
    if (typeof this.data === 'undefined') {
      throw Error('no data secret available');
    }
    const {
      list
    } = this.data;
    const totalItems = list === null || list === void 0 ? void 0 : list.length;
    debug('get full list of packages (%o) has been fetched', totalItems);
    return Promise.resolve(list);
  }
  getPackageStorage(packageName) {
    const packageAccess = (0, _utils.getMatchedPackagesSpec)(packageName, this.config.packages);
    const packagePath = this._getLocalStoragePath(packageAccess ? packageAccess.storage : undefined);
    debug('storage path selected: ', packagePath);
    if (_lodash.default.isString(packagePath) === false) {
      debug('the package %o has no storage defined ', packageName);
      throw _core.errorUtils.getInternalError('storage not found or implemented');
    }
    const packageStoragePath = _path.default.join(
    // FIXME: use getBaseStoragePath instead
    _path.default.resolve(_path.default.dirname(this.config.config_path || ''), packagePath), packageName);
    debug('storage absolute path: ', packageStoragePath);
    return new _localFs.default(packageStoragePath, this.logger);
  }
  async clean() {
    await this._sync();
  }
  _getCustomPackageLocalStorages() {
    const storages = new Map();
    const {
      packages
    } = this.config;
    if (packages) {
      Object.keys(packages || {}).map(pkg => {
        const {
          storage
        } = packages[pkg];
        if (typeof storage === 'string') {
          const storagePath = _path.default.join(this.getStoragePath(), storage);
          debug('add custom storage for %s on %s', storage, storagePath);
          storages.set(storage, storagePath);
        }
      });
    }
    return storages;
  }
  async _sync() {
    debug('sync database started');
    if (this.locked) {
      this.logger.error(ERROR_DB_LOCKED);
      throw new Error(ERROR_DB_LOCKED);
    }
    // Uses sync to prevent ugly race condition
    try {
      const folderName = _path.default.dirname(this.path);
      debug('creating folder %o', folderName);
      await (0, _fs.mkdirPromise)(folderName, {
        recursive: true
      });
      debug('creating folder %o created succeed', folderName);
    } catch (err) {
      this.logger.error({
        err
      }, 'sync create folder has failed with error: @{err}');
      throw err;
    }
    try {
      await (0, _fs.writeFilePromise)(this.path, JSON.stringify(this.data));
      debug('sync write succeeded');
      return null;
    } catch (err) {
      this.logger.error({
        err
      }, 'sync database file failed: @{err}');
      throw err;
    }
  }
  _getLocalStoragePath(storage) {
    const globalConfigStorage = this.getStoragePath();
    if (_lodash.default.isNil(globalConfigStorage)) {
      this.logger.error('property storage in config.yaml is required for using this plugin');
      throw new Error('property storage in config.yaml is required for using this plugin');
    } else {
      if (typeof storage === 'string') {
        return _path.default.join(globalConfigStorage, storage);
      }
      return globalConfigStorage;
    }
  }
  async fetchLocalPackages() {
    try {
      return await (0, _pkgUtils.loadPrivatePackages)(this.path, this.logger);
    } catch (err) {
      // readFileSync is platform specific, macOS, Linux and Windows thrown an error
      // Only recreate if file not found to prevent data loss
      this.logger.info({
        path: this.path
      }, 'no private database found, recreating new one on @{path}');
      if (err.code !== _localFs.noSuchFile) {
        this.locked = true;
        this.logger.error('Failed to read package database file, please check the error printed below:\n', `File Path: ${this.path}\n\n ${err.message}`);
      }
      // if no database is found we set empty placeholders
      return {
        list: [],
        secret: ''
      };
    }
  }
  async getTokenDb() {
    if (!this.tokenDb) {
      debug('token database is not defined');
      let adapter;
      if (process.env.NODE_ENV === 'test') {
        debug('token memory adapter');
        adapter = new _Memory.default('');
      } else {
        debug('token async adapter');
        const pathDb = (0, _utils2._dbGenPath)(TOKEN_DB_NAME, this.config);
        adapter = new _FileAsync.default(pathDb);
      }
      debug('token bd generated');
      this.tokenDb = await (0, _lowdb.default)(adapter);
    }
    return this.tokenDb;
  }
  async saveToken(token) {
    debug('token key %o', token.key);
    const db = await this.getTokenDb();
    const userData = await db.get(token.user).value();
    debug('user data %o', userData);
    if (_lodash.default.isNil(userData)) {
      await db.set(token.user, [token]).write();
      debug('token user %o new database', token.user);
    } else {
      // types does not match with valid implementation
      // @ts-ignore
      await db.get(token.user).push(token).write();
    }
    debug('data %o', await db.getState());
    debug('token saved %o', token.user);
  }
  async deleteToken(user, tokenKey) {
    const db = await this.getTokenDb();
    const userTokens = await db.get(user).value();
    if (_lodash.default.isNil(userTokens)) {
      throw new Error('user not found');
    }
    debug('tokens %o - %o', userTokens, userTokens.length);
    const remainingTokens = userTokens.filter(({
      key
    }) => {
      debug('key %o', key);
      return key !== tokenKey;
    });
    await db.set(user, remainingTokens).write();
    debug('removed tokens key %o', tokenKey);
  }
  async readTokens(filter) {
    const {
      user
    } = filter;
    debug('read tokens with %o', user);
    const db = await this.getTokenDb();
    const tokens = await db.get(user).value();
    return tokens || [];
  }
}
var _default = exports.default = LocalDatabase;
//# sourceMappingURL=local-database.js.map