/// <reference types="node" />
import { Readable, Writable } from 'stream';
import { VerdaccioError, pluginUtils } from '@verdaccio/core';
import { Logger, Manifest } from '@verdaccio/types';
export declare const fileExist = "EEXISTS";
export declare const noSuchFile = "ENOENT";
export declare const resourceNotAvailable = "EAGAIN";
export declare const packageJSONFileName = "package.json";
export type ILocalFSPackageManager = pluginUtils.StorageHandler & {
    path: string;
};
export declare const fSError: (message: string, code?: number) => VerdaccioError;
export declare function renameTmp(src: string, dst: string): Promise<void>;
export default class LocalFS implements ILocalFSPackageManager {
    path: string;
    logger: Logger;
    constructor(path: string, logger: Logger);
    /**
      * This function allows to update the package
      * This function handle the update package logic, for this plugin
      * we need to lock/unlock handlers for thread-safely and then apply
      * the `handleUpdate` and return the result.
      *
      * The lock could fail on several steps so we need to ensure the
      * file does not get locked if the whole process fails.
      *
        Algorithm:
        1. lock package.json for writing
        2. read package.json
        3. apply external update package handler
        4. return manifest (write is being hanlded into the core)
        5. rename package.json.tmp package.json
      * @param {*} packageName
      * The update package handler could be different based
      * on the action and handled into the core.
      * @param {*} handleUpdate
      */
    updatePackage(packageName: string, handleUpdate: (manifest: Manifest) => Promise<Manifest>): Promise<Manifest>;
    deletePackage(packageName: string): Promise<void>;
    removePackage(): Promise<void>;
    /**
     * Verify if the package exists already.
     * @param name package name
     * @returns
     */
    hasPackage(): Promise<boolean>;
    /**
     * Create a package in the local storage, if package already exist fails.
     * @param name package name
     * @param manifest package manifest
     */
    createPackage(name: string, manifest: Manifest): Promise<void>;
    savePackage(name: string, value: Manifest): Promise<void>;
    readPackage(name: string): Promise<Manifest>;
    hasTarball(fileName: string): Promise<boolean>;
    private removeTempFile;
    /**
     * Write a tarball into the storage
     * @param fileName package name
     * @param param1
     * @returns
     */
    writeTarball(fileName: string, { signal }: {
        signal: any;
    }): Promise<Writable>;
    /**
     * Read a tarball from the storage
     * @param tarballName tarball name eg: foo-1.0.0.tgz
     * @param options {signal} abort signal
     * @returns Readable stream
     */
    readTarball(tarballName: string, { signal }: {
        signal: any;
    }): Promise<Readable>;
    private _readStorageFile;
    private _convertToString;
    _getStorage(fileName?: string): string;
    private writeTempFileAndRename;
    private writeFile;
    private checkCreateFolder;
    private _lockAndReadJSON;
    private _unlockJSON;
}
