"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.packageJSONFileName = exports.noSuchFile = exports.fileExist = exports.fSError = exports.default = void 0;
exports.renameTmp = renameTmp;
exports.resourceNotAvailable = void 0;
var _debug = _interopRequireDefault(require("debug"));
var _fs = _interopRequireDefault(require("fs"));
var _path = _interopRequireDefault(require("path"));
var _sanitizeFilename = _interopRequireDefault(require("sanitize-filename"));
var _stream = require("stream");
var _core = require("@verdaccio/core");
var _fileLocking = require("@verdaccio/file-locking");
var _fs2 = require("./fs");
function _interopRequireDefault(e) { return e && e.__esModule ? e : { default: e }; }
/* eslint-disable no-undef */

const fileExist = exports.fileExist = 'EEXISTS';
const noSuchFile = exports.noSuchFile = 'ENOENT';
const resourceNotAvailable = exports.resourceNotAvailable = 'EAGAIN';
const packageJSONFileName = exports.packageJSONFileName = 'package.json';
const debug = (0, _debug.default)('verdaccio:plugin:local-storage:local-fs');
const fSError = function (message, code = 409) {
  const err = _core.errorUtils.getCode(code, message);
  // FIXME: we should return http-status codes here instead, future improvement
  // @ts-ignore
  err.code = message;
  return err;
};
exports.fSError = fSError;
const tempFile = function (str) {
  return `${str}.tmp${String(Math.random()).slice(2)}`;
};
async function renameTmp(src, dst) {
  debug('rename %s to %s', src, dst);
  try {
    await (0, _fs2.renamePromise)(src, dst);
  } catch (err) {
    debug('error rename %s error %s', src, err === null || err === void 0 ? void 0 : err.message);
    await (0, _fs2.unlinkPromise)(src);
  }
}
class LocalFS {
  constructor(path, logger) {
    this.path = path;
    this.logger = logger;
  }

  /**
    * This function allows to update the package
    * This function handle the update package logic, for this plugin
    * we need to lock/unlock handlers for thread-safely and then apply
    * the `handleUpdate` and return the result.
    *
    * The lock could fail on several steps so we need to ensure the
    * file does not get locked if the whole process fails.
    *
      Algorithm:
      1. lock package.json for writing
      2. read package.json
      3. apply external update package handler
      4. return manifest (write is being hanlded into the core)
      5. rename package.json.tmp package.json
    * @param {*} packageName
    * The update package handler could be different based
    * on the action and handled into the core.
    * @param {*} handleUpdate
    */
  async updatePackage(packageName, handleUpdate) {
    // this plugin lock files on write, we handle all possible scenarios
    let locked = false;
    let manifestUpdated;
    try {
      const manifest = await this._lockAndReadJSON(packageJSONFileName);
      locked = true;
      manifestUpdated = await handleUpdate(manifest);
      if (locked) {
        debug('unlock %s', packageJSONFileName);
        await this._unlockJSON(packageJSONFileName);
        this.logger.debug({
          packageName
        }, 'the package @{packageName} has been updated');
        return manifestUpdated;
      } else {
        this.logger.debug({
          packageName
        }, 'the package @{packageName} has been updated');
        return manifestUpdated;
      }
    } catch (err) {
      // we ensure lock the file
      this.logger.error({
        err,
        packageName
      }, 'error @{err.message}  on update package @{packageName}');
      if (locked) {
        // eslint-disable-next-line no-useless-catch
        try {
          await this._unlockJSON(packageJSONFileName);
          // after unlock bubble up error.
          throw err;
        } catch (err) {
          // unlock could fail, we bubble up error
          throw _core.errorUtils.getInternalError('resource temporarily unavailable');
        }
      } else {
        if (err.code === resourceNotAvailable) {
          throw _core.errorUtils.getInternalError('resource temporarily unavailable');
        } else if (err.code === noSuchFile) {
          throw _core.errorUtils.getNotFound();
        } else {
          throw err;
        }
      }
    }
  }
  async deletePackage(packageName) {
    debug('delete a file/package %o', packageName);
    return await (0, _fs2.unlinkPromise)(this._getStorage(packageName));
  }
  async removePackage() {
    debug('remove a package folder %o', this.path);
    await (0, _fs2.rmdirPromise)(this._getStorage('.'));
  }

  /**
   * Verify if the package exists already.
   * @param name package name
   * @returns
   */
  async hasPackage() {
    const pathName = this._getStorage(packageJSONFileName);
    try {
      const stat = await (0, _fs2.statPromise)(pathName);
      return stat.isFile();
    } catch (err) {
      if (err.code === noSuchFile) {
        debug('dir: %o does not exist %s', pathName, err === null || err === void 0 ? void 0 : err.code);
        return false;
      } else {
        this.logger.error('error on verify a package exist %o', err);
        throw _core.errorUtils.getInternalError('error on verify a package exist');
      }
    }
  }

  /**
   * Create a package in the local storage, if package already exist fails.
   * @param name package name
   * @param manifest package manifest
   */
  async createPackage(name, manifest) {
    debug('create a a new package %o', name);
    const pathPackage = this._getStorage(packageJSONFileName);
    try {
      // https://nodejs.org/dist/latest-v17.x/docs/api/fs.html#file-system-flags
      // 'wx': Like 'w' but fails if the path exists
      await (0, _fs2.openPromise)(pathPackage, 'wx');
    } catch (err) {
      // cannot override a pacakge that already exist
      if (err.code === 'EEXIST') {
        debug('file %o cannot be created, it already exists: %o', name);
        throw fSError(fileExist);
      }
    }
    // Create a new file and it´s folder if does not exist previously
    await this.writeFile(pathPackage, this._convertToString(manifest));
  }
  async savePackage(name, value) {
    debug('save a package %o', name);
    await this.writeFile(this._getStorage(packageJSONFileName), this._convertToString(value));
  }
  async readPackage(name) {
    debug('read a package %o', name);
    try {
      const res = await this._readStorageFile(this._getStorage(packageJSONFileName));
      const data = JSON.parse(res.toString('utf8'));
      debug('read storage file %o has succeeded', name);
      return data;
    } catch (err) {
      if (err.code !== noSuchFile) {
        debug('parse error');
        this.logger.error({
          err,
          name
        }, 'error @{err.message}  on parse @{name}');
      }
      throw err;
    }
  }
  async hasTarball(fileName) {
    const pathName = this._getStorage(fileName);
    return new Promise(resolve => {
      (0, _fs2.accessPromise)(pathName).then(() => {
        resolve(true);
      }).catch(() => resolve(false));
    });
  }

  // remove the temporary file
  async removeTempFile(temporalName) {
    debug('remove temporal file %o', temporalName);
    await (0, _fs2.unlinkPromise)(temporalName);
    debug('removed temporal file %o', temporalName);
  }

  /**
   * Write a tarball into the storage
   * @param fileName package name
   * @param param1
   * @returns
   */
  async writeTarball(fileName, {
    signal
  }) {
    debug('write a tarball %o', fileName);
    const pathName = this._getStorage(fileName);
    await this.checkCreateFolder(pathName);
    // create a temporary file to avoid conflicts or prev corruption files
    const temporalName = _path.default.join(this.path, `${fileName}.tmp-${String(Math.random()).replace(/^0\./, '')}`);
    debug('write a temporal name %o', temporalName);
    let opened = false;
    const writeStream = _fs.default.createWriteStream(temporalName);
    writeStream.on('open', () => {
      opened = true;
    });
    writeStream.on('error', async err => {
      if (opened) {
        this.logger.error({
          err: err.message,
          fileName
        }, 'error on open write tarball for @{pkgName}');
        // TODO: maybe add .once
        writeStream.on('close', async () => {
          await this.removeTempFile(temporalName);
        });
      } else {
        this.logger.error({
          err: err.message,
          fileName
        }, 'error a non open write tarball for @{pkgName}');
        await this.removeTempFile(temporalName);
      }
    });

    // the 'close' event is emitted when the stream and any of its
    // underlying resources (a file descriptor, for example) have been closed.
    // TODO: maybe add .once
    writeStream.on('close', async () => {
      try {
        await renameTmp(temporalName, pathName);
      } catch (err) {
        this.logger.error({
          err
        }, 'error on rename temporal file, please report this is a bug @{err}');
      }
    });

    // if upload is aborted, we clean up the temporal file
    signal === null || signal === void 0 || signal.addEventListener('abort', async () => {
      if (opened) {
        // close always happens, even if error
        writeStream.once('close', async () => {
          await this.removeTempFile(temporalName);
        });
      } else {
        await this.removeTempFile(temporalName);
      }
    }, {
      once: true
    });
    return writeStream;
  }

  /**
   * Read a tarball from the storage
   * @param tarballName tarball name eg: foo-1.0.0.tgz
   * @param options {signal} abort signal
   * @returns Readable stream
   */
  async readTarball(tarballName, {
    signal
  }) {
    const pathName = this._getStorage(tarballName);
    debug('read a tarball %o', pathName);
    const readStream = (0, _stream.addAbortSignal)(signal, _fs.default.createReadStream(pathName));
    readStream.on('open', async function (fileDescriptorId) {
      // if abort, the descriptor is null
      debug('file descriptor id %o', fileDescriptorId);
      if (fileDescriptorId) {
        const stats = await (0, _fs2.fstatPromise)(fileDescriptorId);
        debug('file size %o', stats.size);
        readStream.emit('content-length', stats.size);
      }
    });
    readStream.on('error', error => {
      debug('not tarball found %o for %s message %s', pathName, tarballName, error.message);
    });
    return readStream;
  }
  async _readStorageFile(name) {
    debug('reading the file: %o', name);
    try {
      debug('reading the file: %o', name);
      return await (0, _fs2.readFilePromise)(name);
    } catch (err) {
      debug('error reading the file: %o with error %o', name, err.message);
      throw err;
    }
  }
  _convertToString(value) {
    return JSON.stringify(value);
  }
  _getStorage(fileName = '') {
    const storagePath = _path.default.join(this.path, (0, _sanitizeFilename.default)(fileName));
    debug('get storage %s', storagePath);
    return storagePath;
  }
  async writeTempFileAndRename(dest, fileContent) {
    const tempFilePath = tempFile(dest);
    try {
      // write file on temp location
      await this.checkCreateFolder(tempFilePath);
      await (0, _fs2.writeFilePromise)(tempFilePath, fileContent);
      debug('creating a new file:: %o', dest);
      // rename tmp file to original
      await renameTmp(tempFilePath, dest);
    } catch (err) {
      debug('error on write the file: %o', dest);
      throw err;
    }
  }
  async writeFile(destiny, fileContent) {
    try {
      await this.writeTempFileAndRename(destiny, fileContent);
      debug('write file success %s', destiny);
    } catch (err) {
      if (err && err.code === noSuchFile) {
        await this.checkCreateFolder(destiny);
        // we try again create the temp file
        debug('writing a temp file %s', destiny);
        await this.writeTempFileAndRename(destiny, fileContent);
        debug('write file success %s', destiny);
      } else {
        this.logger.error({
          err: err.message
        }, 'error on write file @{err}');
        throw err;
      }
    }
  }
  async checkCreateFolder(pathName) {
    // Check if folder exists and create it if not
    const dir = _path.default.dirname(pathName);
    try {
      await (0, _fs2.accessPromise)(dir);
    } catch (err) {
      if (err.code === noSuchFile) {
        debug('folder does not exist %o', dir);
        await (0, _fs2.mkdirPromise)(dir, {
          recursive: true
        });
        debug('folder %o created', dir);
      } else {
        debug('error on check folder %o', dir);
        throw err;
      }
    }
  }
  async _lockAndReadJSON(name) {
    const fileName = this._getStorage(name);
    debug('lock and read a file %o', fileName);
    try {
      const data = await (0, _fileLocking.readFileNext)(fileName, {
        lock: true,
        parse: true
      });
      return data;
    } catch (err) {
      this.logger.error({
        err
      }, 'error on lock file @{err.message}');
      debug('error on lock and read json for file: %o', name);
      throw err;
    }
  }
  async _unlockJSON(name) {
    await (0, _fileLocking.unlockFileNext)(this._getStorage(name));
  }
}
exports.default = LocalFS;
//# sourceMappingURL=local-fs.js.map